# geneset_network_analysis_tool : Python package

"""Top-level package for geneset_network_analysis_tool."""

__author__ = """Viola Fanfani & Giovanni Stracquadanio"""
__email__ = 'v.fanfani@sms.ed.ac.uk'
__version__ = '3.3.1'

